# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ...core.utils.fallback import unimplemented_func


def _install():
    """Nothing required for installing sklearn."""


def __dir__():  # pragma: no cover
    try:
        import sklearn
    except ImportError:
        raise AttributeError("sklearn is required but not installed.")
    from .mars_adapters import MARS_SKLEARN_ML_CALLABLES

    return list(MARS_SKLEARN_ML_CALLABLES.keys())


def __getattr__(name: str):  # pragma: no cover
    import inspect

    try:
        import sklearn.model_selection as sk_ml
    except ImportError:
        raise AttributeError("sklearn is required but not installed.")
    from .mars_adapters import MARS_SKLEARN_ML_CALLABLES

    if name in MARS_SKLEARN_ML_CALLABLES:
        return MARS_SKLEARN_ML_CALLABLES[name]
    else:
        if not hasattr(sk_ml, name):
            raise AttributeError(name)
        else:
            if inspect.ismethod(getattr(sk_ml, name)):
                return unimplemented_func()
            else:
                raise AttributeError



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from ...._mars.core.entity import OutputType
from ...._mars.serialization.serializables import AnyField
from ...._mars.utils import has_unknown_shape, is_full_slice
from ...operand import DataOperand, DataOperandMixin


class HuggingfaceGetItem(DataOperand, DataOperandMixin):
    hf_getitem_key = AnyField("hf_getitem_key")

    def __call__(self, dataset):
        return self.new_tileable([dataset], **dataset.params)

    @classmethod
    def _gen_copy_output(cls, inp, op: "HuggingfaceGetItem"):
        out_chunks = []
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([chunk], index=chunk.index)
            out_chunks.append(out_chunk)
        return op.copy().new_tileable(op.inputs, chunks=out_chunks)

    @classmethod
    def _gen_empty_output(cls, inp, op: "HuggingfaceGetItem"):
        first_chunk = inp.chunks[0]
        chunk_op = op.copy().reset_key()
        chunk_op.hf_getitem_key = slice(0, 0)
        out_chunk = chunk_op.new_chunk([first_chunk], index=first_chunk.index)
        return op.copy().new_tileable(op.inputs, chunks=[out_chunk])

    @classmethod
    def _find_chunk_index_by_key(cls, chunks, key):
        for idx, chunk in enumerate(chunks):
            if key < 0:  # pragma: no cover
                raise IndexError(f"Input key {key} is out of bound.")
            if key >= chunk.shape[0]:
                key -= chunk.shape[0]
            else:
                return idx, key
        # pragma: no cover
        raise IndexError(f"Input key {key} is out of bound.")

    @classmethod
    def tile(cls, op: "HuggingfaceGetItem"):
        assert len(op.inputs) == 1
        inp = op.inputs[0]

        if isinstance(op.hf_getitem_key, str):
            return cls._gen_copy_output(inp, op)
        elif isinstance(op.hf_getitem_key, int):
            if has_unknown_shape(*op.inputs):
                yield
            index, key = cls._find_chunk_index_by_key(inp.chunks, op.hf_getitem_key)
            chunk = inp.chunks[index]
            chunk_op = op.copy().reset_key()
            chunk_op.hf_getitem_key = key
            out_chunk = chunk_op.new_chunk([chunk], index=chunk.index)
            return op.copy().new_tileable(op.inputs, chunks=[out_chunk])
        elif isinstance(op.hf_getitem_key, slice):
            if is_full_slice(op.hf_getitem_key):
                return cls._gen_copy_output(inp, op)
            else:
                start = op.hf_getitem_key.start
                stop = op.hf_getitem_key.stop
                assert op.hf_getitem_key.step is None
                if start >= stop:
                    # For empty slice, e.g. s[3:1], s[3:3], we translate the
                    # execution to the first chunk[0:0].
                    return cls._gen_empty_output(inp, op)
                else:
                    if has_unknown_shape(*op.inputs):
                        yield
                    try:
                        start_index, start_key = cls._find_chunk_index_by_key(
                            inp.chunks, start
                        )
                    except IndexError:
                        return cls._gen_empty_output(inp, op)
                    try:
                        stop_index, stop_key = cls._find_chunk_index_by_key(
                            inp.chunks, stop
                        )
                    except IndexError:
                        stop_index = len(inp.chunks) - 1
                        stop_key = None
                    chunks = []
                    for index, chunk in enumerate(inp.chunks):
                        if start_index <= index <= stop_index:
                            chunk_op = op.copy().reset_key()
                            slice_start = start_key if index == start_index else None
                            slice_stop = stop_key if index == stop_index else None
                            chunk_op.hf_getitem_key = slice(
                                slice_start, slice_stop, None
                            )
                            out_chunk = chunk_op.new_chunk([chunk], index=chunk.index)
                            chunks.append(out_chunk)
                        elif index > stop_index:
                            break
                    return op.copy().new_tileable(op.inputs, chunks=chunks)
        else:  # pragma: no cover
            raise NotImplementedError(
                f"Not support getitem with key type: {type(op.hf_getitem_key)}"
            )

    @classmethod
    def execute(cls, ctx, op: "HuggingfaceGetItem"):
        inp = ctx[op.inputs[0].key]
        out = op.outputs[0]
        ctx[out.key] = inp.__getitem__(op.hf_getitem_key)


def getitem(dataset, key: Union[int, slice, str]):
    if not isinstance(key, (str, int, slice)):
        raise NotImplementedError(f"Not support getitem with key type: {type(key)}")
    if isinstance(key, slice) and key.step is not None:
        raise NotImplementedError(f"Not support getitem with slice and step: {key}")
    op = HuggingfaceGetItem(output_types=[OutputType.object], hf_getitem_key=key)
    return op(dataset).execute().fetch()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import BinaryIO, Dict, Iterator, List, TextIO, Tuple, Union
from urllib.parse import ParseResult, urlparse, urlunparse

from fsspec import filesystem
from fsspec.core import stringify_path

from ...utils import implements
from .core import FileSystem, path_type


class FsSpecAdapter(FileSystem):
    def __init__(self, scheme: str, **kwargs):
        self._fs = filesystem(scheme, **kwargs)
        self._scheme = scheme

    @implements(FileSystem.cat)
    def cat(self, path: path_type) -> bytes:
        return self._fs.cat_file(self._normalize_path(path))

    @implements(FileSystem.ls)
    def ls(self, path: path_type) -> List[path_type]:
        if self.isfile(path):  # pragma: no cover
            return self._append_scheme(
                self._fs.ls(self._normalize_path(path), detail=False)
            )
        entries = []
        for entry in self._fs.ls(self._normalize_path(path), detail=False):
            if entry.strip("/") in path.strip("/"):  # pragma: no cover
                continue
            if isinstance(entry, Dict):
                entries.append(entry.get("name"))
            elif isinstance(entry, str):
                entries.append(entry)
            else:  # pragma: no cover
                raise TypeError(f"Expect str or dict, but got {type(entry)}")
        return self._append_scheme(entries)

    @implements(FileSystem.delete)
    def delete(self, path: path_type, recursive: bool = False):
        raise NotImplementedError

    @implements(FileSystem.stat)
    def stat(self, path: path_type) -> Dict:
        return self._fs.info(self._normalize_path(path))

    @implements(FileSystem.rename)
    def rename(self, path: path_type, new_path: path_type):
        raise NotImplementedError

    @implements(FileSystem.mkdir)
    def mkdir(self, path: path_type, create_parents: bool = True):
        raise NotImplementedError

    @implements(FileSystem.exists)
    def exists(self, path: path_type):
        return self._fs.exists(self._normalize_path(path))

    @implements(FileSystem.isdir)
    def isdir(self, path: path_type) -> bool:
        return self._fs.isdir(self._normalize_path(path))

    @implements(FileSystem.isfile)
    def isfile(self, path: path_type) -> bool:
        return self._fs.isfile(self._normalize_path(path))

    @implements(FileSystem._isfilestore)
    def _isfilestore(self) -> bool:
        raise NotImplementedError

    @implements(FileSystem.open)
    def open(self, path: path_type, mode: str = "rb") -> Union[BinaryIO, TextIO]:
        return self._fs.open(self._normalize_path(path), mode=mode)

    @implements(FileSystem.walk)
    def walk(self, path: path_type) -> Iterator[Tuple[str, List[str], List[str]]]:
        for root, dirs, files in self._fs.walk(path):
            yield self._append_scheme([root])[0], self._append_scheme(
                dirs
            ), self._append_scheme(files)

    @implements(FileSystem.glob)
    def glob(self, path: path_type, recursive: bool = False) -> List[path_type]:
        from ._glob import FileSystemGlob

        return self._append_scheme(
            FileSystemGlob(self).glob(self._normalize_path(path), recursive=recursive)
        )

    @staticmethod
    def _normalize_path(path: path_type) -> str:
        """
        Stringify path and remove its scheme.
        """
        path_str = stringify_path(path)
        parsed = urlparse(path_str)
        if parsed.scheme:
            return urlunparse(
                ParseResult(
                    scheme="",
                    netloc=parsed.netloc,
                    path=parsed.path,
                    params="",
                    query="",
                    fragment="",
                )
            )
        else:
            return path_str

    def _append_scheme(self, paths: List[path_type]) -> List[path_type]:
        return [
            urlunparse(
                ParseResult(
                    scheme=self._scheme,
                    netloc="",
                    path=path,
                    params="",
                    query="",
                    fragment="",
                )
            )
            for path in paths
        ]



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, TupleField
from ..array_utils import device, get_array_module
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin


class TensorBroadcastTo(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.BROADCAST_TO

    _input = KeyField("input")
    _shape = TupleField("shape")

    def __init__(self, shape=None, **kw):
        super().__init__(_shape=shape, **kw)

    @property
    def shape(self):
        return self._shape

    def __call__(self, tensor, shape):
        return self.new_tensor([tensor], shape)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        in_tensor = op.inputs[0]
        shape = op.shape
        new_dim = tensor.ndim - in_tensor.ndim

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_shape = shape[:new_dim] + tuple(
                s if in_tensor.shape[idx] != 1 else shape[new_dim + idx]
                for idx, s in enumerate(c.shape)
            )
            chunk_idx = (0,) * new_dim + c.index
            chunk_op = op.copy().reset_key()
            chunk_op._shape = chunk_shape
            out_chunk = chunk_op.new_chunk(
                [c], shape=chunk_shape, index=chunk_idx, order=tensor.order
            )
            out_chunks.append(out_chunk)

        nsplits = [
            tuple(
                c.shape[i]
                for c in out_chunks
                if all(idx == 0 for j, idx in enumerate(c.index) if j != i)
            )
            for i in range(len(out_chunks[0].shape))
        ]
        new_op = op.copy()
        return new_op.new_tensors(
            [in_tensor],
            tensor.shape,
            order=tensor.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        xp = get_array_module(ctx[op.input.key])
        input_data = ctx[op.input.key]
        device_id = input_data.device.id if hasattr(input_data, "device") else -1

        with device(device_id):
            shape = op.shape
            if any(np.isnan(s) for s in shape):
                shape = list(shape)
                new_dim = len(shape) - input_data.ndim
                for i in range(input_data.ndim):
                    if np.isnan(shape[i + new_dim]):
                        shape[i + new_dim] = input_data.shape[i]
            ctx[op.outputs[0].key] = xp.broadcast_to(input_data, shape)


def broadcast_to(tensor, shape):
    """Broadcast an tensor to a new shape.

    Parameters
    ----------
    tensor : array_like
        The tensor to broadcast.
    shape : tuple
        The shape of the desired array.

    Returns
    -------
    broadcast : Tensor

    Raises
    ------
    ValueError
        If the tensor is not compatible with the new shape according to Mars's
        broadcasting rules.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([1, 2, 3])
    >>> mt.broadcast_to(x, (3, 3)).execute()
    array([[1, 2, 3],
           [1, 2, 3],
           [1, 2, 3]])
    """
    from ..core import Tensor

    tensor = tensor if isinstance(tensor, Tensor) else astensor(tensor)
    shape = tuple(shape) if isinstance(shape, (list, tuple)) else (shape,)

    if any(np.isnan(s) for s in tensor.shape):
        raise ValueError(
            "input tensor has unknown shape, need to call `.execute()` first"
        )

    if tensor.shape == shape:
        return tensor

    new_ndim = len(shape) - tensor.ndim
    if new_ndim < 0:
        raise ValueError(
            "input operand has more dimensions than allowed by the axis remapping"
        )
    if any(o != n for o, n in zip(tensor.shape, shape[new_ndim:]) if o != 1):
        raise ValueError(
            "operands could not be broadcast together "
            f"with remapped shapes [original->remapped]: {tensor.shape} "
            f"and requested shape {shape}"
        )

    op = TensorBroadcastTo(shape, dtype=tensor.dtype, sparse=tensor.issparse())
    return op(tensor, shape)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import concurrent.futures
import os
import time

import psutil
import pytest
from xoscar.aio.lru import clear_all_alru_caches
from xoscar.backends.router import Router

from .config import option_context
from .core.mode import is_build_mode, is_kernel_mode
from .utils import lazy_import

ray = lazy_import("ray")
MARS_CI_BACKEND = os.environ.get("MARS_CI_BACKEND", "mars")


@pytest.fixture(autouse=True)
def auto_cleanup(request):
    request.addfinalizer(clear_all_alru_caches)


@pytest.fixture(scope="module", autouse=True)
def check_router_cleaned(request):
    def route_checker():
        if Router.get_instance() is not None:
            assert len(Router.get_instance()._mapping) == 0
            assert len(Router.get_instance()._local_mapping) == 0

    request.addfinalizer(route_checker)


@pytest.fixture
def stop_mars():
    try:
        yield
    finally:
        import xorbits._mars

        xorbits._mars.stop_server()


@pytest.fixture(scope="module")
def _new_test_session(check_router_cleaned):
    from .deploy.oscar.tests.session import new_test_session

    sess = new_test_session(
        address="test://127.0.0.1",
        backend=MARS_CI_BACKEND,
        init_local=True,
        default=True,
        timeout=300,
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)
            Router.set_instance(None)


@pytest.fixture(scope="module")
def _new_integrated_test_session(check_router_cleaned):
    from .deploy.oscar.tests.session import new_test_session

    sess = None
    for i in range(3):
        try:
            sess = new_test_session(
                address="127.0.0.1",
                backend=MARS_CI_BACKEND,
                init_local=True,
                n_worker=2,
                default=True,
                timeout=300,
            )
        except ChildProcessError:
            time.sleep(1)
            if i == 2:
                raise
            else:
                continue
        else:
            break
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            try:
                sess.stop_server(isolation=False)
            except concurrent.futures.TimeoutError:
                Router.set_instance(None)
                subprocesses = psutil.Process().children(recursive=True)
                for proc in subprocesses:
                    proc.terminate()
                for proc in subprocesses:
                    try:
                        proc.wait(1)
                    except (psutil.TimeoutExpired, psutil.NoSuchProcess):
                        pass
                    try:
                        proc.kill()
                    except psutil.NoSuchProcess:
                        pass


@pytest.fixture(scope="module")
def _new_gpu_test_session(check_router_cleaned):  # pragma: no cover
    from .deploy.oscar.tests.session import new_test_session
    from .resource import cuda_count

    cuda_devices = list(range(min(cuda_count(), 2)))

    sess = new_test_session(
        address="127.0.0.1",
        backend=MARS_CI_BACKEND,
        init_local=True,
        n_worker=1,
        n_cpu=1,
        cuda_devices=cuda_devices,
        default=True,
        timeout=300,
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)
            Router.set_instance(None)


@pytest.fixture
def setup(_new_test_session):
    _new_test_session.as_default()
    yield _new_test_session
    assert not (is_build_mode() or is_kernel_mode())


@pytest.fixture
def setup_cluster(_new_integrated_test_session):
    _new_integrated_test_session.as_default()
    yield _new_integrated_test_session


@pytest.fixture
def setup_gpu(_new_gpu_test_session):  # pragma: no cover
    _new_gpu_test_session.as_default()
    yield _new_test_session



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.adapter import (
    mars_flatiter,
    register_converter,
    to_mars,
    wrap_mars_callable,
)


@register_converter(from_cls_list=[mars_flatiter])
class FlatIter:
    def __init__(self, mars_obj: "mars_flatiter"):
        self._mars_obj = mars_obj

    def __getitem__(self, item):
        return wrap_mars_callable(
            self._mars_obj.__getitem__, attach_docstring=False, is_cls_member=True
        )(item)

    def __setitem__(self, key, value):
        self._mars_obj[key] = to_mars(value)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional

from .juicefs.config import (
    PersistentVolumeClaimConfig,
    PersistentVolumeConfig,
    SecretConfig,
)

try:
    from kubernetes import client as kube_client
    from kubernetes.client import ApiClient
except ImportError:  # pragma: no cover
    client = None
    ApiClient = None


class ExternalStorage(ABC):
    def __init__(
        self,
        namespace: Optional[str],
        api_client: ApiClient,
    ):
        self._namespace = namespace
        self._api_client = api_client
        self._core_api = kube_client.CoreV1Api(self._api_client)

    @abstractmethod
    def build(self):
        """
        Build the external storage
        """


class JuicefsK8SStorage(ExternalStorage):
    def __init__(
        self,
        namespace: Optional[str],
        api_client: ApiClient,
        external_storage_config: Optional[dict],
    ):
        super().__init__(namespace=namespace, api_client=api_client)
        self._external_storage_config = external_storage_config

    def build(self):
        """
        Create pv, secret, and pvc
        """
        secret_config = SecretConfig(
            metadata_url=self._external_storage_config["metadata_url"],
            bucket=self._external_storage_config["bucket"],
        )

        persistent_volume_config = PersistentVolumeConfig(namespace=self._namespace)

        persistent_volume_claim_config = PersistentVolumeClaimConfig(
            namespace=self._namespace
        )
        self._core_api.create_namespaced_secret(self._namespace, secret_config.build())
        self._core_api.create_persistent_volume(persistent_volume_config.build())
        self._core_api.create_namespaced_persistent_volume_claim(
            self._namespace, persistent_volume_claim_config.build()
        )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np

from ... import remote as mr
from ... import tensor as mt
from ...tensor.datasource import tensor as astensor


def instance_softmax_loss_and_sgd(W, X, y, loss_reg_W, i):
    X = X[i].reshape((1, X[i].shape[0]))
    y = y[i]

    N, D = X.shape
    K = W.shape[1]

    y_obs = mt.eye(K)[y]

    exp_X_W_over_sum = mt.exp(X @ W) / mt.sum(mt.exp(X @ W), axis=1).reshape(-1, 1)

    loss = -1 * mt.sum(y_obs * mt.log(exp_X_W_over_sum)) + loss_reg_W

    dW = mt.zeros(shape=(D, K))

    # Matrix approach
    dW = -1 * X.T @ (y_obs - (exp_X_W_over_sum))

    return loss, dW


def gradient_descent(
    X,
    y,
    learning_rate=1e-3,
    reg=1e-5,
    max_epochs=100,
    batch_size=20,
    fit_intercept=True,
    verbose=0,
):
    # assume y takes values 0...K-1 where K is number of classes
    num_classes = (mt.max(y) + 1).to_numpy()

    num_train, dim = X.shape
    num_iters_per_epoch = int(math.floor(1.0 * num_train / batch_size))

    # need extra entries if fit_intercept
    if fit_intercept:
        X = mt.hstack((X, mt.ones((num_train, 1))))
        W = 0.001 * mt.random.randn(dim + 1, num_classes).execute()
    else:
        X = astensor(X)
        W = 0.001 * mt.random.randn(dim, num_classes).execute()

    X = X.execute()
    y = y.execute()

    for _ in range(max_epochs):
        # perform mini-batch SGD update
        perm_idx = np.random.permutation(num_train)
        for it in range(num_iters_per_epoch):
            idx = perm_idx[it * batch_size : (it + 1) * batch_size]
            cur_W = W.execute()
            loss_reg_W = (0.5 * reg * mt.sum(mt.square(W))).execute()
            dW_reg_W = (reg * cur_W).execute()

            funcs = [
                mr.spawn(
                    instance_softmax_loss_and_sgd,
                    args=(
                        cur_W,
                        X,
                        y,
                        loss_reg_W,
                        i,
                    ),
                )
                for i in idx
            ]

            outs = mr.ExecutableTuple(funcs).execute().fetch()
            grad_tensors = [out[1] for out in outs]

            # update parameters
            W = (
                cur_W
                - learning_rate * (mt.sum(grad_tensors, axis=0) / len(idx) + dW_reg_W)
            ).execute()

    return W



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Set

from .....dataframe.core import BaseDataFrameData


def get_cols_exclude_index(inp: BaseDataFrameData, cols: Any) -> Set[Any]:
    ret = set()
    if isinstance(cols, (list, tuple)):
        for col in cols:
            if col in inp.dtypes.index:
                # exclude index
                ret.add(col)
    else:
        if cols in inp.dtypes.index:
            # exclude index
            ret.add(cols)
    return ret



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from collections import defaultdict
from enum import Enum
from itertools import count
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Type

from pandas.api.types import is_float_dtype, is_integer_dtype

from ..utils import safe_repr_str

if TYPE_CHECKING:  # pragma: no cover
    from ..core.adapter import MarsEntity

import numpy as np


class DataType(Enum):
    object_ = 1
    scalar = 2
    tensor = 3
    dataframe = 4
    series = 5
    index = 6
    categorical = 7
    dataframe_groupby = 8
    series_groupby = 9
    dataset = 10


DATA_MEMBERS: Dict[DataType, Dict[str, Any]] = defaultdict(dict)


class AutoConversionType(Enum):
    int_conversion = 1
    float_conversion = 2

    def convert(self, val):
        if self == AutoConversionType.int_conversion:
            return int(val)
        else:
            return float(val)

    def generate_error_msg(self, val):
        if self == AutoConversionType.int_conversion:
            return f"{val} object cannot be interpreted as an integer."
        else:
            return f"{val} object cannot be interpreted as a float."


class Data:
    data_type: DataType

    __fields: List[str] = [
        "data_type",
        "_mars_entity",
    ]

    def __init__(self, *args, **kwargs):
        self.data_type: DataType = kwargs.pop("data_type")
        self._mars_entity = kwargs.pop("mars_entity", None)
        if len(args) > 0 or len(kwargs) > 0:  # pragma: no cover
            raise TypeError(f"Unexpected args {args} or kwargs {kwargs}.")

    @classmethod
    def from_mars(cls, mars_entity: "MarsEntity") -> "Data":
        from ..core.adapter import (
            MARS_CATEGORICAL_TYPE,
            MARS_DATAFRAME_GROUPBY_TYPE,
            MARS_DATAFRAME_TYPE,
            MARS_INDEX_TYPE,
            MARS_OBJECT_TYPE,
            MARS_SERIES_GROUPBY_TYPE,
            MARS_SERIES_TYPE,
            MARS_TENSOR_TYPE,
        )

        # import MARS_DATASET_TYPE from datasets instead of core.adapter
        # to avoid recursive import.
        from ..datasets import MARS_DATASET_TYPE

        if isinstance(mars_entity, MARS_DATAFRAME_TYPE):
            data_type = DataType.dataframe
        elif isinstance(mars_entity, MARS_SERIES_TYPE):
            data_type = DataType.series
        elif isinstance(mars_entity, MARS_DATAFRAME_GROUPBY_TYPE):
            data_type = DataType.dataframe_groupby
        elif isinstance(mars_entity, MARS_SERIES_GROUPBY_TYPE):
            data_type = DataType.series_groupby
        elif isinstance(mars_entity, MARS_TENSOR_TYPE):
            data_type = DataType.tensor
        elif isinstance(mars_entity, MARS_INDEX_TYPE):
            data_type = DataType.index
        elif isinstance(mars_entity, MARS_CATEGORICAL_TYPE):
            data_type = DataType.categorical
        elif isinstance(mars_entity, MARS_DATASET_TYPE):
            data_type = DataType.dataset
        elif isinstance(mars_entity, MARS_OBJECT_TYPE):
            data_type = DataType.object_
        else:
            raise NotImplementedError(f"Unsupported mars type {type(mars_entity)}")
        return Data(mars_entity=mars_entity, data_type=data_type)

    def __setattr__(self, key: str, value: Any):
        from .adapter import MemberProxy

        if key in self.__fields:
            object.__setattr__(self, key, value)
        else:
            return MemberProxy.setattr(self._mars_entity, key, value)

    def __str__(self):
        if self._mars_entity is not None:
            return self._mars_entity.__str__()
        else:  # pragma: no cover
            return super().__str__()

    def __repr__(self):
        if self._mars_entity is not None:
            return self._mars_entity.__repr__()
        else:  # pragma: no cover
            return super().__repr__()

    def __array__(self):
        if self._mars_entity is not None:
            return self._mars_entity.__array__()
        else:  # pragma: no cover
            raise AttributeError("__array__")


class DataRefMeta(type):
    """
    Used to bind methods to subclasses of DataRef according to their data type.
    """

    __cls_members: Dict[str, Any]

    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        if name == "DataFrame":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.dataframe])
        elif name == "Series":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.series])
        elif name == "Index":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.index])
        elif name == "DataFrameGroupBy":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.dataframe_groupby])
        elif name == "SeriesGroupBy":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.series_groupby])
        elif name == "ndarray":
            setattr(cls, "__cls_members", DATA_MEMBERS[DataType.tensor])
        return cls

    def __getattr__(cls, item: str):
        members = object.__getattribute__(cls, "__cls_members")
        if item not in members:
            raise AttributeError(item)
        else:
            return members[item]

    def __instancecheck__(cls: Type, instance: Any) -> bool:
        if not issubclass(instance.__class__, DataRef):
            # not a DataRef instance.
            return False

        if cls is DataRef:
            # isinstance(x, DataRef).
            return cls in instance.__class__.__mro__
        else:
            # for subclass like isinstance(x, DataFrame),
            # check its data_type if match with cls
            data_type = instance.data.data_type
            try:
                return data_type == SUB_CLASS_TO_DATA_TYPE[cls]
            except KeyError:
                # subclassing DataRef subclasses is not allowed.
                raise TypeError(f"Illegal subclass {instance.__class__.__name__}")


class DataRef(metaclass=DataRefMeta):
    data: Data

    __fields = ["data"]

    def __dir__(self) -> Iterable[str]:
        return list(DATA_MEMBERS[self.data.data_type])

    def __init__(self, data: Data):
        self.data = data

    def __getattr__(self, item):
        from .adapter import MemberProxy

        return MemberProxy.getattr(self, item)

    def __setattr__(self, key, value):
        try:
            if key in self.__fields:
                object.__setattr__(self, key, value)
            else:
                self.data.__setattr__(key, value)
        except AttributeError:
            if key in self.dtypes:
                self.__setitem__(key, value)
            else:
                warnings.warn(
                    "xorbits.pandas doesn't allow columns to be created via a new attribute name.",
                    UserWarning,
                )
                object.__setattr__(self, key, value)

    def _own_data(self):
        from .adapter import own_data

        return own_data(self.data._mars_entity)

    def __iter__(self):
        # Mars entity hasn't implemented __iter__, however `iter(mars_entity)`
        # still works, it's because iteration is supported by `__getitem__` that
        # accepts integers 0,1,.., it can be seen as a "legacy feature" that not
        # recommended. Here we implement __iter__ for some data types, others keep
        # behaviors with Mars.
        if self._own_data():
            # if own data, return iteration on data directly
            yield from iter(self.data._mars_entity.op.data)
        else:
            if self.data.data_type == DataType.dataframe:
                yield from iter(self.data._mars_entity.dtypes.index)
            elif self.data.data_type == DataType.series:
                for batch_data in self.data._mars_entity.iterbatch():
                    yield from batch_data.__iter__()
            elif self.data.data_type == DataType.index:
                for batch_data in self.data._mars_entity.to_series().iterbatch():
                    yield from batch_data.__iter__()
            else:

                def gen():
                    counter = count()
                    while True:
                        try:
                            yield self.__getitem__(next(counter))
                        except IndexError:
                            break

                yield from gen()

    def __len__(self):
        from .._mars.core import HasShapeTileable
        from .execution import run

        if isinstance(self.data._mars_entity, HasShapeTileable):
            try:
                return int(self.data._mars_entity.shape[0])
            except (IndexError, ValueError):
                # shape is unknown, execute it
                run(self)
                try:
                    return int(self.data._mars_entity.shape[0])
                except (IndexError, ValueError):  # happens when dimension is 0
                    raise TypeError(
                        f"object with shape {self.data._mars_entity.shape} has no len()"
                    )

        else:
            raise TypeError(f"object of type '{self.data.data_type}' has no len()")

    @property
    def shape(self):
        from .._mars.core import HasShapeTileable
        from .execution import run

        if isinstance(self.data._mars_entity, HasShapeTileable):
            if np.isnan(self.data._mars_entity.shape).any():
                run(self)
            return self.data._mars_entity.shape
        else:
            raise TypeError(f"object of type '{self.data.data_type}' has no shape")

    @safe_repr_str
    def __str__(self):
        from .execution import run

        if self._own_data():
            # skip execution if own data
            return self.data._mars_entity.op.data.__str__()
        else:
            run(self)
            return self.data.__str__()

    @safe_repr_str
    def __repr__(self):
        from .execution import run

        if self._own_data():
            # skip execution if own data
            return self.data._mars_entity.op.data.__repr__()
        else:
            run(self)
            return self.data.__repr__()

    def __array__(self):
        from .execution import run

        if self._own_data():
            return self.data._mars_entity.op.data.__array__()
        else:
            run(self)
            return self.data.__array__()

    def _to_int_or_float(self, conversion: AutoConversionType, cast: bool = False):
        from .execution import run

        data_type = self.data.data_type
        if (
            data_type == DataType.tensor
            and len(self.shape) == 0
            and (is_integer_dtype(self.dtype) or is_float_dtype(self.dtype))
        ):
            run(self)
            return conversion.convert(self.to_numpy())
        elif data_type == DataType.object_:
            run(self)
            data_object = self.to_object()
            if cast or isinstance(data_object, int) or isinstance(data_object, float):
                return conversion.convert(self.to_object())
            else:
                raise TypeError(conversion.generate_error_msg(self.data.data_type))
        else:
            raise TypeError(conversion.generate_error_msg(self.data.data_type))

    def __bool__(self):
        data_type = self.data.data_type
        from .execution import run

        if data_type == DataType.tensor:
            if len(self.shape) == 0:
                run(self)
                return bool(self.to_numpy())
            else:
                if self.__len__() <= 1:
                    run(self)
                    return bool(self.to_numpy())
                else:
                    raise ValueError(
                        f"ValueError: The truth value of a {data_type} with more than one element is ambiguous. "
                        f"Use a.any() or a.all()"
                    )
        elif data_type == DataType.object_:
            run(self)
            return bool(self.to_object())
        elif (
            data_type == DataType.dataframe
            or data_type == DataType.series
            or data_type == DataType.index
        ):
            raise ValueError(
                f"The truth value of a {data_type} is ambiguous. Use a.empty, a.bool(), a.item(), a.any() or a.all()."
            )
        else:
            raise ValueError(f"{data_type} cannot be converted to boolean values.")

    def __int__(self):
        return self._to_int_or_float(AutoConversionType.int_conversion, cast=True)

    def __float__(self):
        return self._to_int_or_float(AutoConversionType.float_conversion, cast=True)

    def __index__(self):
        try:
            val = self._to_int_or_float(AutoConversionType.int_conversion, cast=False)
            return val
        except TypeError:
            raise TypeError(
                f"{self.data.data_type} object cannot be interpreted as an integer."
            )


SUB_CLASS_TO_DATA_TYPE: Dict[Type[DataRef], DataType] = dict()


def register_cls_to_type(data_type: DataType) -> Callable:
    def _wrap_cls(cls: Type[DataRef]):
        SUB_CLASS_TO_DATA_TYPE[cls] = data_type
        return cls

    return _wrap_cls



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import AnyField
from .core import DataFrameReductionMixin, DataFrameReductionOperand


class DataFrameCustomReduction(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.CUSTOM_REDUCTION
    _func_name = "custom_reduction"

    custom_reduction = AnyField("custom_reduction")

    @property
    def is_atomic(self):
        return True

    def get_reduction_args(self, axis=None):
        return dict()


def build_custom_reduction_result(df, custom_reduction_obj, method=None):
    output_type = OutputType.series if df.ndim == 2 else OutputType.scalar
    op = DataFrameCustomReduction(
        custom_reduction=custom_reduction_obj,
        output_types=[output_type],
        method=method,
    )
    return op(df)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Type

from ....core import OperandType, TileableGraph
from ..core import (
    OperandBasedOptimizationRule,
    OptimizationRecords,
    OptimizationRule,
    Optimizer,
)


class TileableOptimizer(Optimizer):
    """
    Tileable Optimizer
    """


def register_optimization_rule():
    def wrap(rule_type: Type[OptimizationRule]):
        TileableOptimizer.register_rule(rule_type)

    return wrap


def register_operand_based_optimization_rule(op_types: List[Type[OperandType]]):
    def wrap(rule_type: Type[OperandBasedOptimizationRule]):
        for op_type in op_types:
            rule_type.register_operand(op_type)
        TileableOptimizer.register_rule(rule_type)

    return wrap


def optimize(tileable_graph: TileableGraph) -> OptimizationRecords:
    return TileableOptimizer.optimize(tileable_graph)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from typing import Optional

from ...config import KubeConfig


class SecretConfig(KubeConfig):
    """
    Configuration builder for Juicefs secret
    """

    _default_kind = "Secret"
    api_version = "v1"

    def __init__(
        self,
        metadata_url: str,
        bucket: str,
        kind: Optional[str] = None,
    ):
        self._kind = kind or self._default_kind
        self._metadata_url = metadata_url
        self._bucket = bucket

    def build(self):
        return {
            "kind": self._kind,
            "metadata": {"name": "juicefs-secret"},
            "type": "Opaque",
            "stringData": {
                "name": "jfs",
                "metaurl": self._metadata_url,
                "storage": "file",
                "bucket": self._bucket,
            },
        }


class PersistentVolumeConfig(KubeConfig):
    """
    Juicefs persistent volume builder
    """

    _default_kind = "PersistentVolume"
    api_version = "v1"

    def __init__(
        self,
        namespace: str,
        kind: Optional[str] = None,
    ):
        self._namespace = namespace
        self._kind = kind or self._default_kind

    def build(self):
        return {
            "kind": self._kind,
            "metadata": {
                "name": f"juicefs-pv-{self._namespace}",
                "labels": {"juicefs-name": f"juicefs-fs-{self._namespace}"},
            },
            "spec": {
                "capacity": {
                    "storage": "200M"
                },  # For now, JuiceFS CSI Driver doesn't support setting storage capacity. Fill in any valid string is fine. ( Reference: https://juicefs.com/docs/csi/guide/pv/ )
                "volumeMode": "Filesystem",
                "mountOptions": [
                    "subdir=/data/{ns}/{id}".format(
                        ns=self._namespace, id=uuid.uuid4().hex
                    )
                ],  # Mount in sub directory to achieve data isolation. See https://juicefs.com/docs/csi/guide/pv/#create-storage-class for more references.
                "accessModes": [
                    "ReadWriteMany"
                ],  # accessModes is restricted to ReadWriteMany because it's the most suitable mode for our system. See https://kubernetes.io/docs/concepts/storage/persistent-volumes/#access-modes for more reference
                "persistentVolumeReclaimPolicy": "Retain",  # persistentVolumeReclaimPolicy is restricted to Retain for Static provisioning. See https://juicefs.com/docs/csi/guide/resource-optimization/#reclaim-policy for more references.
                "csi": {
                    "driver": "csi.juicefs.com",
                    "volumeHandle": f"juicefs-pv-{self._namespace}",
                    "fsType": "juicefs",
                    "nodePublishSecretRef": {
                        "name": "juicefs-secret",
                        "namespace": self._namespace,
                    },
                },
            },
        }


class PersistentVolumeClaimConfig(KubeConfig):
    """
    Juicefs persistent volume claim builder
    """

    _default_kind = "PersistentVolumeClaim"
    api_version = "v1"

    def __init__(
        self,
        namespace: str,
        kind: Optional[str] = None,
    ):
        self._namespace = namespace
        self._kind = kind or self._default_kind

    def build(self):
        return {
            "kind": self._kind,
            "metadata": {"name": "juicefs-pvc", "namespace": self._namespace},
            "spec": {
                "accessModes": ["ReadWriteMany"],
                "volumeMode": "Filesystem",
                "storageClassName": "",
                "resources": {"requests": {"storage": "200M"}},
                "selector": {
                    "matchLabels": {"juicefs-name": f"juicefs-fs-{self._namespace}"}
                },
            },
        }



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import warnings
from types import ModuleType

from ...core.adapter import MarsOutputType, mars_remote


def unimplemented_func():
    """
    Not implemented yet.
    """
    raise NotImplementedError(f"This function is not implemented yet.")


def wrap_fallback_module_method(
    mod: ModuleType, func_name: str, output_type: MarsOutputType, warning_str: str
):
    # wrap member function
    @functools.wraps(getattr(mod, func_name))
    def _wrapped(*args, **kwargs):
        warnings.warn(warning_str, RuntimeWarning)

        # use mars remote to execute functions
        def execute_func(f_name: str, *args, **kwargs):
            from xorbits.core.adapter import MarsEntity

            def _replace_data(nested):
                if isinstance(nested, dict):
                    vals = list(nested.values())
                else:
                    vals = list(nested)

                new_vals = []
                for val in vals:
                    if isinstance(val, (dict, list, tuple, set)):
                        new_val = _replace_data(val)
                    else:
                        if isinstance(val, MarsEntity):
                            new_val = val.fetch()
                        else:
                            new_val = val
                    new_vals.append(new_val)
                if isinstance(nested, dict):
                    return type(nested)((k, v) for k, v in zip(nested.keys(), new_vals))
                else:
                    return type(nested)(new_vals)

            return getattr(mod, f_name)(*_replace_data(args), **_replace_data(kwargs))

        new_args = (func_name,) + args
        ret = mars_remote.spawn(
            execute_func, args=new_args, kwargs=kwargs, output_types=output_type
        )
        if output_type == MarsOutputType.df_or_series:
            ret = ret.ensure_data()
        else:
            ret = ret.execute()
        return ret

    return _wrapped



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..utils import implement_scipy, infer_dtype
from .core import TensorTupleOp, _register_special_op


@_register_special_op
class TensorAiry(TensorTupleOp):
    _func_name = "airy"
    _n_outputs = 4


@implement_scipy(spspecial.airy)
@infer_dtype(spspecial.airy, multi_outputs=True)
def airy(z, out=None, **kwargs):
    op = TensorAiry(**kwargs)
    return op(z, out=out)


@_register_special_op
class TensorAirye(TensorTupleOp):
    _func_name = "airye"
    _n_outputs = 4


@implement_scipy(spspecial.airye)
@infer_dtype(spspecial.airye, multi_outputs=True)
def airye(z, out=None, **kwargs):
    op = TensorAirye(**kwargs)
    return op(z, out=out)


@_register_special_op
class TensorItairy(TensorTupleOp):
    _func_name = "itairy"
    _n_outputs = 4


@implement_scipy(spspecial.itairy)
@infer_dtype(spspecial.itairy, multi_outputs=True)
def itairy(x, out=None, **kwargs):
    op = TensorItairy(**kwargs)
    return op(x, out=out)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ...core import OutputType, get_output_types, recursive_tile
from ...serialization.serializables import DictField, IndexField, StringField
from ..core import IndexValue
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_concatenated_rows_frame, parse_index


class GroupByNthSelector(DataFrameOperand, DataFrameOperandMixin):
    _op_module_ = "dataframe.groupby"

    groupby_params = DictField("groupby_params")
    n = IndexField("n")
    dropna = StringField("dropna", default=None)

    def __call__(self, groupby):
        df = groupby
        while df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            df = df.inputs[0]
        selection = groupby.op.groupby_params.pop("selection", None)
        if df.ndim > 1 and selection:
            if isinstance(selection, tuple) and selection not in df.dtypes:
                selection = list(selection)

            result_df = df[selection]
        else:
            result_df = df

        self._output_types = (
            [OutputType.dataframe] if result_df.ndim == 2 else [OutputType.series]
        )
        params = result_df.params
        params["shape"] = (np.nan,) + result_df.shape[1:]
        if isinstance(df.index_value.value, IndexValue.RangeIndex):
            params["index_value"] = parse_index(pd.RangeIndex(-1), df.key)

        return self.new_tileable([df], **params)

    @classmethod
    def tile(cls, op: "GroupByNthSelector"):
        in_df = op.inputs[0]
        groupby_params = op.groupby_params.copy()
        selection = groupby_params.pop("selection", None)
        if len(in_df.shape) > 1:
            in_df = build_concatenated_rows_frame(in_df)
        out_df = op.outputs[0]
        # if there is only one chunk, tile with a single chunk
        if len(in_df.chunks) <= 1:
            new_shape = (np.nan,)
            new_nsplits = ((np.nan,),)
            if out_df.ndim > 1:
                new_shape += (out_df.shape[1],)
                new_nsplits += ((out_df.shape[1],),)
            c = in_df.chunks[0]
            chunk_op = op.copy().reset_key()
            params = out_df.params
            params["shape"] = new_shape
            params["index"] = (0,) * out_df.ndim
            out_chunk = chunk_op.new_chunk([c], **params)

            tileable_op = op.copy().reset_key()
            return tileable_op.new_tileables(
                [in_df], nsplits=new_nsplits, chunks=[out_chunk], **params
            )

        if in_df.ndim > 1 and selection:
            if isinstance(selection, tuple) and selection not in in_df.dtypes:
                selection = list(selection)

            if not isinstance(selection, list):
                pre_selection = [selection]
            else:
                pre_selection = list(selection)

            if isinstance(groupby_params.get("by"), list):
                pre_selection += [
                    el for el in groupby_params["by"] if el not in pre_selection
                ]

            if len(pre_selection) != in_df.shape[1]:
                in_df = yield from recursive_tile(in_df[pre_selection])

        # pre chunks
        pre_chunks = []
        for c in in_df.chunks:
            pre_op = op.copy().reset_key()
            pre_op._output_types = get_output_types(c)
            pre_op.groupby_params = op.groupby_params.copy()
            pre_op.groupby_params.pop("selection", None)
            params = c.params
            params["shape"] = (np.nan,) + c.shape[1:]
            pre_chunks.append(pre_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        new_op._output_types = get_output_types(in_df)
        new_nsplits = ((np.nan,) * len(in_df.nsplits[0]),) + in_df.nsplits[1:]
        pre_tiled = new_op.new_tileable(
            [in_df], chunks=pre_chunks, nsplits=new_nsplits, **in_df.params
        )
        # generate groupby
        grouped = yield from recursive_tile(pre_tiled.groupby(**groupby_params))
        if selection:
            grouped = yield from recursive_tile(grouped[selection])

        # generate post chunks
        post_chunks = []
        for c in grouped.chunks:
            post_op = op.copy().reset_key()
            post_op.groupby_params = op.groupby_params.copy()
            post_op.groupby_params.pop("selection", None)
            if op.output_types[0] == OutputType.dataframe:
                index = c.index
            else:
                index = (c.index[0],)
            params = out_df.params
            params["index"] = index
            post_chunks.append(post_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        new_nsplits = ((np.nan,) * len(in_df.nsplits[0]),)
        if out_df.ndim > 1:
            new_nsplits += ((out_df.shape[1],),)
        return new_op.new_tileables(
            [in_df], chunks=post_chunks, nsplits=new_nsplits, **out_df.params
        )

    @classmethod
    def execute(cls, ctx, op: "GroupByNthSelector"):
        in_data = ctx[op.inputs[0].key]
        params = op.groupby_params.copy()
        selection = params.pop("selection", None)

        if hasattr(in_data, "groupby"):
            grouped = in_data.groupby(**params)
        else:
            grouped = in_data
        if selection:
            grouped = grouped[selection]
        result = grouped.nth(op.n, op.dropna)
        ctx[op.outputs[0].key] = result


def nth(groupby, n, dropna=None):
    """
    Take the nth row from each group if n is an int, or a subset of rows
    if n is a list of ints.

    If dropna, will take the nth non-null row, dropna is either
    Truthy (if a Series) or 'all', 'any' (if a DataFrame);
    this is equivalent to calling dropna(how=dropna) before the
    groupby.

    Parameters
    ----------
    n : int or list of ints
        a single nth value for the row or a list of nth values
    dropna : None or str, optional
        apply the specified dropna operation before counting which row is
        the nth row. Needs to be None, 'any' or 'all'

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': [1, 1, 2, 1, 2],
    ...                    'B': [np.nan, 2, 3, 4, 5]}, columns=['A', 'B'])
    >>> g = df.groupby('A')
    >>> g.nth(0).execute()
            B
    A
    1  NaN
    2  3.0
    >>> g.nth(1).execute()
            B
    A
    1  2.0
    2  5.0
    >>> g.nth(-1).execute()
            B
    A
    1  4.0
    2  5.0
    >>> g.nth([0, 1]).execute()
            B
    A
    1  NaN
    1  2.0
    2  3.0
    2  5.0

    Specifying ``dropna`` allows count ignoring NaN

    >>> g.nth(0, dropna='any').execute()
            B
    A
    1  2.0
    2  3.0

    NaNs denote group exhausted when using dropna

    >>> g.nth(3, dropna='any').execute()
        B
    A
    1 NaN
    2 NaN

    Specifying ``as_index=False`` in ``groupby`` keeps the original index.

    >>> df.groupby('A', as_index=False).nth(1).execute()
        A    B
    1  1  2.0
    4  2  5.0
    """
    groupby_params = groupby.op.groupby_params.copy()
    groupby_params.pop("as_index", None)
    op = GroupByNthSelector(n=n, dropna=dropna, groupby_params=groupby_params)
    return op(groupby)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from ..core import MarsRequestHandler


class MarsApiEntryHandler(MarsRequestHandler):
    def get(self):
        import xorbits._mars

        version = xorbits._mars.__version__
        self.write(json.dumps({"mars_version": version}))


web_handlers = {"/api": MarsApiEntryHandler}



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
from collections import defaultdict
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from ... import opcodes, options
from ...core import OutputType
from ...core.context import Context, get_context
from ...core.operand import MapReduceOperand, OperandStage
from ...serialization.serializables import (
    BoolField,
    Int8Field,
    Int32Field,
    Int64Field,
    StringField,
)
from ...utils import estimate_pandas_size
from ..groupby.aggregation import SizeRecorder
from ..operands import DataFrameOperandMixin, DataFrameShuffleProxy
from ..utils import build_concatenated_rows_frame, hash_dataframe_on, parse_index


class DataFrameNunique(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.NUNIQUE

    axis = Int8Field("axis")
    dropna = BoolField("dropna")
    shuffle_size = Int32Field("shuffle_size")
    combine_size = Int64Field("combine_size")
    method = StringField("method")
    execution_stage = StringField("execution_stage", default=None)
    size_recorder_name = StringField("size_recorder_name")

    def __init__(self, **kw):
        super().__init__(**kw)

    def __call__(self, df: Union["pd.DataFrame", "pd.Series"]):
        if df.ndim == 2:
            shape = df.shape
            if shape[0] == 0:
                res_dtype = np.dtype(np.float64)
            else:
                res_dtype = np.dtype(np.int64)

            if self.axis == 0:
                res_shape = (len(df.columns),)
                res_index_value = df.columns
            else:
                res_shape = (len(df),)
                res_index_value = df.index

            return self.new_series(
                [df],
                shape=res_shape,
                dtype=res_dtype,
                index_value=parse_index(res_index_value),
                name=None,
            )
        else:
            return self.new_scalar([df], dtype=np.dtype(np.int64))

    @classmethod
    def _gen_chunk_params(cls, in_df: Union["pd.DataFrame", "pd.Series"]):
        output_type = OutputType.dataframe if in_df.ndim == 2 else OutputType.series
        shape = (np.nan, np.nan) if in_df.ndim == 2 else (np.nan,)
        index = (0, 0) if in_df.ndim == 2 else (0,)
        return output_type, shape, index

    @classmethod
    def _gen_tree_pre_chunks(
        cls,
        input_chunks: List,
        output_type: OutputType,
        shape: Tuple,
        op: "DataFrameNunique",
    ):
        chunks = []
        for c in input_chunks:
            _new_op = op.copy().reset_key()
            _new_op.execution_stage = "pre"
            _new_op.output_types = [output_type]
            chunks.append(
                _new_op.new_chunk(
                    inputs=[c],
                    shape=shape,
                    index=c.index,
                )
            )
        return chunks

    @classmethod
    def _gen_tree_agg_chunks(
        cls,
        chunks: List,
        output_type: OutputType,
        shape: Tuple,
        index: Tuple,
        op: "DataFrameNunique",
    ):
        combine_size = op.combine_size
        while len(chunks) > combine_size:
            new_chunks = []
            for i in range(0, len(chunks), combine_size):
                chks = chunks[i : i + combine_size]
                if len(chks) == 1:
                    chk = chks[0]
                else:
                    union_op = op.copy().reset_key()
                    union_op.output_types = [output_type]
                    union_op.execution_stage = "agg"
                    for j, c in enumerate(chks):
                        c._index = (j, 0)
                    chk = union_op.new_chunk(chks, shape=shape, index=index)
                new_chunks.append(chk)
            chunks = new_chunks

        if len(chunks) > 1:
            union_op = op.copy().reset_key()
            union_op.output_types = [output_type]
            union_op.execution_stage = "agg"
            agg_chunk = union_op.new_chunk(chunks, shape=shape, index=index)
        else:
            agg_chunk = chunks[0]
        return agg_chunk

    @classmethod
    def _tile_tree(
        cls, in_df: Union["pd.DataFrame", "pd.Series"], op: "DataFrameNunique"
    ):
        output_type, shape, index = cls._gen_chunk_params(in_df)
        pre_chunks = cls._gen_tree_pre_chunks(in_df.chunks, output_type, shape, op)
        agg_chunk = cls._gen_tree_agg_chunks(pre_chunks, output_type, shape, index, op)

        result_chunks = []
        new_op = op.copy().reset_key()
        new_op.execution_stage = "post"
        if in_df.ndim == 2:
            shape = (np.nan,)
            index_value = (
                parse_index(in_df.columns) if op.axis == 0 else parse_index(in_df.index)
            )
            params = dict(
                shape=shape,
                index=(0,),
                dtype=np.dtype(np.int64),
                index_value=index_value,
            )
            result_chunks.append(new_op.new_chunk([agg_chunk], **params))

            _new_op = op.copy()
            params = op.outputs[0].params.copy()
            params["nsplits"] = ((np.nan,) * len(result_chunks),)
            params["chunks"] = result_chunks
            return _new_op.new_seriess(op.inputs, **params)
        else:
            dtype = np.dtype(np.int64)
            params = dict(index=(0,), dtype=dtype, shape=())
            result_chunks.append(new_op.new_chunk([agg_chunk], **params))

            _new_op = op.copy()
            return _new_op.new_scalars(op.inputs, chunks=result_chunks, dtype=dtype)

    @classmethod
    def _tile_one_chunk_df(cls, in_df: "pd.DataFrame", op: "DataFrameNunique"):
        result_chunks = []
        new_op = op.copy().reset_key()
        shape = (np.nan,)
        index_value = (
            parse_index(in_df.columns) if op.axis == 0 else parse_index(in_df.index)
        )
        params = dict(
            shape=shape, index=(0,), dtype=np.dtype(np.int64), index_value=index_value
        )
        result_chunks.append(new_op.new_chunk(in_df.chunks, **params))

        _new_op = op.copy()
        params = op.outputs[0].params.copy()
        params["nsplits"] = ((np.nan,) * len(result_chunks),)
        params["chunks"] = result_chunks
        return _new_op.new_seriess(op.inputs, **params)

    @classmethod
    def _tile_one_chunk_series(cls, in_df: "pd.Series", op: "DataFrameNunique"):
        result_chunks = []
        new_op = op.copy().reset_key()
        dtype = np.dtype(np.int64)
        params = dict(index=(0,), dtype=dtype, shape=())
        result_chunks.append(new_op.new_chunk(in_df.chunks, **params))

        _new_op = op.copy()
        return _new_op.new_scalars(op.inputs, chunks=result_chunks, dtype=dtype)

    @classmethod
    def tile_one_chunk(
        cls, in_df: Union["pd.DataFrame", "pd.Series"], op: "DataFrameNunique"
    ):
        if in_df.ndim == 2:
            return cls._tile_one_chunk_df(in_df, op)
        else:
            return cls._tile_one_chunk_series(in_df, op)

    @classmethod
    def _tile_series_shuffle(cls, in_df: "pd.Series", op: "DataFrameNunique"):
        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op.stage = OperandStage.map
            map_op.shuffle_size = len(in_df.chunks)
            map_op._output_types = [OutputType.series]
            map_chunks.append(
                map_op.new_chunk(
                    [chunk],
                    shape=(np.nan,),
                    index=chunk.index,
                )
            )
        proxy_chunk = DataFrameShuffleProxy(output_types=[OutputType.series]).new_chunk(
            map_chunks, shape=()
        )

        # generate reduce chunks
        reduce_chunks = []
        for chunk in in_df.chunks:
            reduce_op = op.copy().reset_key()
            reduce_op._output_types = [OutputType.series]
            reduce_op.stage = OperandStage.reduce
            reduce_op.n_reducers = len(in_df.chunks)
            reduce_chunks.append(
                reduce_op.new_chunk([proxy_chunk], shape=(np.nan,), index=chunk.index)
            )

        out_chunks = []
        for chunk in reduce_chunks:
            new_op = op.copy().reset_key()
            params = dict(shape=(), index=(chunk.index[0],), dtype=np.dtype(np.int64))
            out_chunks.append(new_op.new_chunk([chunk], **params))

        combine_chunks = []
        combine_op = op.copy().reset_key()
        combine_op.stage = OperandStage.combine
        params = dict(shape=(), index=(0,), dtype=np.dtype(np.int64))
        combine_chunks.append(combine_op.new_chunk(out_chunks, **params))

        new_op = op.copy()
        return new_op.new_scalars(
            op.inputs, chunks=combine_chunks, dtype=np.dtype(np.int64)
        )

    @classmethod
    def _tile_df_shuffle(cls, in_df: "pd.DataFrame", op: "DataFrameNunique"):
        map_output_types = (
            [OutputType.series] if op.axis == 0 else [OutputType.dataframe]
        )
        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op.stage = OperandStage.map
            map_op.shuffle_size = len(in_df.chunks)
            map_op._output_types = map_output_types
            map_shape = (np.nan,) if op.axis == 0 else (np.nan, np.nan)
            chunk_inputs = [chunk]
            map_chunks.append(
                map_op.new_chunk(
                    chunk_inputs,
                    shape=map_shape,
                    index=chunk.index,
                )
            )
        proxy_chunk = DataFrameShuffleProxy(output_types=map_output_types).new_chunk(
            map_chunks, shape=()
        )

        # generate reduce chunks
        reduce_chunks = []
        for chunk in in_df.chunks:
            reduce_op = op.copy().reset_key()
            reduce_op._output_types = [OutputType.dataframe]
            reduce_op.stage = OperandStage.reduce
            reduce_op.n_reducers = len(in_df.chunks)
            reduce_chunks.append(
                reduce_op.new_chunk(
                    [proxy_chunk], shape=(np.nan, np.nan), index=chunk.index
                )
            )

        out_chunks = []
        for chunk in reduce_chunks:
            new_op = op.copy().reset_key()
            new_shape = (np.nan,)

            params = dict(
                shape=new_shape,
                index=(chunk.index[0],),
                dtype=np.dtype(np.int64),
                index_value=parse_index(None),
            )
            out_chunks.append(new_op.new_chunk([chunk], **params))

        combine_chunks = []
        combine_op = op.copy().reset_key()
        combine_shape = (np.nan,)
        combine_op.stage = OperandStage.combine
        params = dict(
            shape=combine_shape,
            index=(0,),
            dtype=np.dtype(np.int64),
            index_value=parse_index(None),
        )
        combine_chunks.append(combine_op.new_chunk(out_chunks, **params))

        new_op = op.copy()
        params = op.outputs[0].params.copy()
        params["nsplits"] = ((np.nan,) * len(combine_chunks),)
        params["chunks"] = combine_chunks
        return new_op.new_seriess(op.inputs, **params)

    @classmethod
    def _tile_shuffle(
        cls, in_df: Union["pd.DataFrame", "pd.Series"], op: "DataFrameNunique"
    ):
        if in_df.ndim == 1:
            return cls._tile_series_shuffle(in_df, op)
        else:
            return cls._tile_df_shuffle(in_df, op)

    @classmethod
    def _tile_auto(
        cls, in_df: Union["pd.DataFrame", "pd.Series"], op: "DataFrameNunique"
    ):
        output_type, shape, _ = cls._gen_chunk_params(in_df)

        ctx = get_context()
        combine_size = op.combine_size
        size_recorder_name = str(uuid.uuid4())
        size_recorder = ctx.create_remote_object(size_recorder_name, SizeRecorder)

        # collect the first pre chunk, run it get the size before and after pre stage
        pre_chunks = cls._gen_tree_pre_chunks(
            in_df.chunks[:combine_size], output_type, shape, op
        )
        for c in pre_chunks:
            c.op.size_recorder_name = size_recorder_name

        # yield chunks with source chunks to avoid submitting chunks repeatedly issue
        yield in_df.chunks[:combine_size] + pre_chunks

        raw_size, agg_size = size_recorder.get()
        # destroy size recorder
        ctx.destroy_remote_object(size_recorder_name)

        # here's a simple rule for deciding tree or shuffle, may improve it later
        if sum(agg_size) / sum(raw_size) > 0.1:
            return cls._tile_shuffle(in_df, op)
        else:
            return cls._tile_tree(in_df, op)

    @classmethod
    def tile(cls, op: "DataFrameNunique"):
        in_df = op.inputs[0]
        if in_df.ndim == 2:
            in_df = build_concatenated_rows_frame(in_df)
        if len(in_df.chunks) == 1:
            return cls.tile_one_chunk(in_df, op)
        if op.method == "auto":
            if len(in_df.chunks) <= op.combine_size:
                return cls._tile_tree(in_df, op)
            else:
                return (yield from cls._tile_auto(in_df, op))
        elif op.method == "tree":
            return cls._tile_tree(in_df, op)
        else:
            return cls._tile_shuffle(in_df, op)

    @classmethod
    def execute_map(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        chunk = op.outputs[0]
        input = ctx[op.inputs[0].key]

        if input.ndim == 1:
            data = input.drop_duplicates()
            filters = hash_dataframe_on(data, data.index, op.shuffle_size)
            for index_idx, index_filter in enumerate(filters):
                reducer_index = (index_idx,)
                ctx[chunk.key, reducer_index] = (
                    ctx.get_current_chunk().index,
                    data.iloc[index_filter],
                )
        else:
            if op.axis == 0:
                reducer_index_to_series_list = defaultdict(list)
                for col, col_data in input.items():
                    drop_duplicates_data = col_data.drop_duplicates()
                    filters = hash_dataframe_on(
                        drop_duplicates_data,
                        drop_duplicates_data.index,
                        op.shuffle_size,
                    )
                    for index_idx, index_filter in enumerate(filters):
                        reducer_index = (index_idx, chunk.index[1])
                        part_data = drop_duplicates_data.iloc[index_filter]
                        reducer_index_to_series_list[reducer_index].append(part_data)
                for reducer_index, data in reducer_index_to_series_list.items():
                    ctx[chunk.key, reducer_index] = (
                        ctx.get_current_chunk().index,
                        data,
                    )

            else:
                unique_input = cls._drop_duplicates_by_row(input)
                # df that contains list element cannot be hashed,
                # so here we just mock a ``filters`` object,
                # since how to split is not important in ``axis=1`` case.
                mock_hash_series = pd.Series(range(len(unique_input)))
                mock_hash_series.index = unique_input.index
                idx_to_grouped = pd.RangeIndex(0, len(unique_input)).groupby(
                    mock_hash_series % op.shuffle_size
                )
                filters = [
                    idx_to_grouped.get(i, pd.Index([])) for i in range(op.shuffle_size)
                ]
                for index_idx, index_filter in enumerate(filters):
                    reducer_index = (index_idx, chunk.index[1])
                    ctx[chunk.key, reducer_index] = (
                        ctx.get_current_chunk().index,
                        unique_input.iloc[index_filter],
                    )

    @classmethod
    def execute_combine(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        out = op.outputs[0]
        inputs = [ctx[i.key] for i in op.inputs]

        data = inputs[0]
        if isinstance(data, int):
            ctx[out.key] = sum(inputs)
        else:
            for s in inputs[1:]:
                data = data.add(s, fill_value=0)
            ctx[out.key] = data.astype(np.int64)

    @classmethod
    def execute_reduce(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        chunk = op.outputs[0]
        input_idx_to_df = dict(op.iter_mapper_data(ctx))

        row_idxes = sorted(input_idx_to_df.keys())

        if op.axis is None or op.axis == 1:
            res = []
            for row_idx in row_idxes:
                row = input_idx_to_df.get(row_idx, None)
                if row is not None:
                    res.append(row)

            data = pd.concat(res, axis=0)
            ctx[chunk.key] = data.drop_duplicates() if op.axis is None else data
        else:
            column_to_series = {}
            for row_idx in row_idxes:
                series_list = input_idx_to_df.get(row_idx, None)
                if series_list is not None:
                    for series in series_list:
                        series_data = series.reset_index(drop=True)
                        if len(series_data) == 0:
                            continue
                        if series.name not in column_to_series and len(series_data) > 0:
                            column_to_series[series.name] = [series_data]
                        else:
                            column_to_series[series.name].append(series_data)

            if column_to_series:
                column_to_na_values = {}
                for name in column_to_series.keys():
                    col_value = (
                        pd.concat(column_to_series[name], axis=0)
                        .reset_index(drop=True)
                        .drop_duplicates()
                    )
                    col_value_has_na: bool = col_value.isnull().any()
                    if not col_value_has_na:
                        column_to_na_values[name] = col_value[0]
                    column_to_series[name] = col_value
                df = pd.concat(column_to_series.values(), axis=1).fillna(
                    value=column_to_na_values
                )
                ctx[chunk.key] = df
            else:
                ctx[chunk.key] = pd.DataFrame()

    @classmethod
    def _drop_duplicates_by_column(cls, df: "pd.DataFrame") -> "pd.DataFrame":
        data = []
        for col, col_data in df.items():
            data.append(col_data.drop_duplicates().tolist())
        res = pd.DataFrame([data])
        res.columns = df.columns
        return res

    @classmethod
    def _drop_duplicates_by_row(cls, df: "pd.DataFrame") -> "pd.DataFrame":
        res = pd.DataFrame(columns=[0])
        for i, row in df.iterrows():
            data = row.drop_duplicates().tolist()
            res.loc[i] = [data]
        return res

    @classmethod
    def execute_pre(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        input = ctx[op.inputs[0].key]
        chunk = op.outputs[0]
        if input.ndim == 1:
            res = input.drop_duplicates()
            ctx[chunk.key] = res
        else:
            if op.axis == 0:
                res = cls._drop_duplicates_by_column(input)
                ctx[chunk.key] = res
            else:
                res = cls._drop_duplicates_by_row(input)
                ctx[chunk.key] = res

        if getattr(op, "size_recorder_name", None) is not None:
            # record_size
            raw_size = estimate_pandas_size(input)
            # when agg by a list of methods, agg_size should be sum
            agg_size = estimate_pandas_size(res)
            size_recorder = ctx.get_remote_object(op.size_recorder_name)
            size_recorder.record(raw_size, agg_size)

    @classmethod
    def execute_agg(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        inputs = [ctx[inp.key] for inp in op.inputs]
        data = pd.concat(inputs, axis=0)
        if op.axis is None:
            res = data.drop_duplicates()
            ctx[op.outputs[0].key] = res
        else:
            if op.axis == 0:
                res = []
                for col, col_data in data.items():
                    res.append(col_data.explode().drop_duplicates().tolist())
                res = pd.DataFrame([res])
                res.columns = data.columns
                ctx[op.outputs[0].key] = res
            else:
                res = pd.DataFrame(columns=[0])
                for i, row in data.iterrows():
                    res.loc[i] = [row.explode().drop_duplicates().tolist()]
                ctx[op.outputs[0].key] = res

    @classmethod
    def _gen_post_result(
        cls, input: "pd.DataFrame", op: "DataFrameNunique"
    ) -> "pd.Series":
        res = []
        for i, row in input.iterrows():
            res.append(row.explode().nunique(dropna=op.dropna))
        res = pd.Series(res)
        res.index = input.index
        return res

    @classmethod
    def execute_post(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        input = ctx[op.inputs[0].key]
        if input.ndim == 1:
            ctx[op.outputs[0].key] = input.nunique(dropna=op.dropna)
        else:
            if op.axis == 0:
                res = []
                for col, col_data in input.items():
                    res.append(col_data.explode().nunique(dropna=op.dropna))
                res = pd.Series(res)
                res.index = input.columns
                ctx[op.outputs[0].key] = res
            else:
                res = cls._gen_post_result(input, op)
                ctx[op.outputs[0].key] = res

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameNunique"):
        if op.execution_stage is not None:
            if op.execution_stage == "pre":
                cls.execute_pre(ctx, op)
            elif op.execution_stage == "agg":
                cls.execute_agg(ctx, op)
            else:
                cls.execute_post(ctx, op)
        else:
            if op.stage == OperandStage.map:
                cls.execute_map(ctx, op)
            elif op.stage == OperandStage.reduce:
                cls.execute_reduce(ctx, op)
            elif op.stage == OperandStage.combine:
                cls.execute_combine(ctx, op)
            else:
                chunk = op.outputs[0]
                df = ctx[op.inputs[0].key]
                if df.ndim == 2:
                    if op.axis == 0:
                        ctx[chunk.key] = df.nunique(axis=op.axis, dropna=op.dropna)
                    else:
                        ctx[chunk.key] = (
                            df.nunique(axis=op.axis, dropna=op.dropna)
                            if len(df) == 0
                            else cls._gen_post_result(df, op)
                        )
                else:
                    ctx[chunk.key] = df.nunique(dropna=op.dropna)


def nunique_dataframe(
    df,
    axis: Union[int, str] = 0,
    dropna: bool = True,
    method: str = "auto",
    combine_size: Optional[int] = None,
):
    """
    Count distinct observations over requested axis.

    Return Series with number of distinct observations. Can ignore NaN
    values.

    Parameters
    ----------
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to use. 0 or 'index' for row-wise, 1 or 'columns' for
        column-wise.
    dropna : bool, default True
        Don't include NaN in the counts.
    method : str, default 'auto'
        execute via tree or shuffle way.
    combine_size : int, default None
        combine chunk sizes when executing via tree way.

    Returns
    -------
    Series

    See Also
    --------
    Series.nunique: Method nunique for Series.
    DataFrame.count: Count non-NA cells for each column or row.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': [1, 2, 3], 'B': [1, 1, 1]})
    >>> df.nunique().execute()
    A    3
    B    1
    dtype: int64

    >>> df.nunique(axis=1).execute()
    0    1
    1    2
    2    2
    dtype: int64
    """
    if not (axis in (0, "index", 1, "columns")):
        raise ValueError(f"No axis named {axis} for object type DataFrame")
    if method not in ("tree", "shuffle", "auto"):
        raise ValueError(f"the method input `{method}` is not allowed.")
    if axis == "index":
        axis = 0
    elif axis == "columns":
        axis = 1

    if combine_size is None:
        combine_size = options.combine_size

    op = DataFrameNunique(
        axis=axis, dropna=dropna, method=method, combine_size=combine_size
    )
    return op(df)


def nunique_series(
    series,
    dropna: bool = True,
    method: str = "auto",
    combine_size: Optional[int] = None,
):
    """
    Return number of unique elements in the object.

    Excludes NA values by default.

    Parameters
    ----------
    dropna : bool, default True
        Don't include NaN in the count.
    method : str, default 'auto'
        execute via tree or shuffle way.
    combine_size : int, default None
        combine chunk sizes when executing via tree way.

    Returns
    -------
    int

    See Also
    --------
    DataFrame.nunique: Method nunique for DataFrame.
    Series.count: Count non-NA/null observations in the Series.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 3, 5, 7, 7])
    >>> s.execute()
    0    1
    1    3
    2    5
    3    7
    4    7
    dtype: int64

    >>> s.nunique().execute()
    4
    """
    if method not in ("tree", "shuffle", "auto"):
        raise ValueError(f"the method input `{method}` is not allowed.")

    if combine_size is None:
        combine_size = options.combine_size

    op = DataFrameNunique(
        axis=None, dropna=dropna, method=method, combine_size=combine_size
    )
    return op(series)
